// Name: fractal.cpp
// Written by Grant Macklem
//		Grant.Macklem@Colorado.EDU
// 	2:00 PM Recitation
//		TA: Steve Bjorg
// Last Modified on Wed, Dec 09, 1998 at 2:04 PM
/*******************************************************************************
* fractal.exe  A fractal generator based upon the collision of a charged
*              particle in an electric field.
* Copyright (C) 1998 Grant Macklem
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* You may contact me, the author by emailing Grant.Macklem@Colorado.EDU
*******************************************************************************/

#include "winbgim.h"
#include <math.h>
#include <iostream.h>

const int WIDTH = 400;
const int HEIGHT = 400;
const int XBORDER = 3;
const int YBORDER = 3;
/*
const int WIDTH = 900;
const int HEIGHT = 675;
const int XBORDER = 6.75;
const int YBORDER = 5.0625;
*/
const double E_MASS = 1;					// Mass of an electron
const double CHARGE_FACTOR = 1;			// Strength of electric force
const int C_RADIUS = 15;					// Radius of point charges in pixels

const int LIMIT = 6;							// Speed limit for moving particles
const int MAX_PARTICLES = 50;				// Maximum number of point charges

/************************ Structures ******************************************/
struct Particle
{
	double x;			// Current x position of the particle
   double y;			// Current y position of the particle
   int charge;			// Charge of the particle (+1 or -1, depending)
   double sx;			// Currents x-speed of the particle
   double sy;			// Current y-speed of the particle
   int color;			// Color which the particle (or point) is drawn
   double mass;		// Mass of the particle

};

/******************************** Prototypes **********************************/
int ytranslate
	(double y);						// Cartesian Coordinate to be translated into pixels

int xtranslate
	(double x);						// Cartesian Coordinate to be translated into pixels

double xpixel_to_cart(int x);
double ypixel_to_cart(int x);

// Processes each pixel on the screen and determines which point charge it was
// finally attracted to.  Then, it assigns a color to the pixel.
void get_color(Particle& pixel, int i, int j, int many_particles, Particle other[ ]);

// Puts the default values for each element of the particle
Particle assign_defaults(int i, int j);

// Runs the particle around in the charge field
void process_particle(Particle& pixel, Particle charge[ ], int many_particles);

// Gets the location and charge of all the other point charges in the
// electrical field.  Return value is how many point charges inputted.
int get_point_charges(Particle other[ ]);

// Checks to see if the particle intersected a charge and either moves it or
// calculates the color of intersection
void move_particle(Particle& pixel, Particle charge[ ], int many_particles);

// Calculates the distance from the center of the charge along a perpendicular
// line to the last portion of the particle's trajectory.
double get_distance(Particle pixel, Particle other);


/******************************************************************************/
// Global Variable(s)
bool intersect = FALSE;			// Whether or not the particle intersected the charge
int color = 0;
int iterations = 0;				// Number of iterations to go through

/******************************** FUNCTIONS ***********************************/
void main ()
{
	int many_particles;
	int i, j;
   Particle pixel;							// Variable for current processing particle
   Particle other[MAX_PARTICLES];		// Array holding point charge information

   // Input point charges
   many_particles = get_point_charges(other);

	init_window(WIDTH, HEIGHT);
	for (int l = 0; l < many_particles; l++)
		circle (xtranslate(other[l].x), ytranslate(other[l].y), 15);

   for (i = 0; i < WIDTH; i++)
   {
   	for (j = 0; j < HEIGHT; j++)
      {
      	get_color(pixel, i, j, many_particles, other);
         putpixel(i, j, pixel.color);
         intersect = FALSE;
			color = 0;
      }
   }

   while(!kbhit());
}


int get_point_charges(Particle other[ ])
{
	int many_particles = 0;
   int tempx, tempy;			// Variables to hold the user's inputted data
	cout << "Electric Field Fractal Program" << endl;
   cout << "Copyright (C) Fall 1998 Grant Macklem" << endl;
   cout << "University of Colorado at Boulder" << endl << endl;
   cout << "Input the point charge locations in pixels" << endl;
	cout << "Enter -1 for x to stop" << endl;
   cout << "Enter the maximum number of iterations on each point: ";
   cin >> iterations;

   do
   {
   	// input particle here
      cout << "x: ";
      cin >> tempx;
      if (tempx == -1)
      	return many_particles;

      other[many_particles].x = xpixel_to_cart(tempx);

      cout << "y: ";
      cin >> tempy;
      other[many_particles].y = ypixel_to_cart(tempy);

      cout << "Concentration of Charge (for example, +15 for 15 positrons or -15 for 15" << endl;
		cout << "electons: ";
      cin >> other[many_particles].charge;

      other[many_particles].sx = 0;
		other[many_particles].sy = 0;
      other[many_particles].color = many_particles + 1;

      many_particles++;
   }  while (many_particles < MAX_PARTICLES);
   return many_particles;
}


void get_color(Particle& pixel, int i, int j, int many_particles, Particle other[ ])
{
	pixel = assign_defaults(i, j);

	do
   {
		process_particle(pixel, other, many_particles);
		move_particle(pixel, other, many_particles);
   }	while (intersect == FALSE);
}

void move_particle(Particle& pixel, Particle charge[ ], int many_particles)
{
	double newx = pixel.x + pixel.sx;
   double newy = pixel.y + pixel.sy;
   double dist;		// Perpendicular distance (squared) from center of charge to line
   double left, top, right, bottom;		// Box coordinates that point charge is in
   double c_radius = xpixel_to_cart(WIDTH/2 + C_RADIUS);
   int l;
   Particle other;

	for (l = 0; l < many_particles; l++)
	{
   	other = charge[l];
      intersect = TRUE;

		color ++;
      if (color > iterations)
      {	// Particle went into orbit
			pixel.color = 0;
         return;
      }
		// Coordinates of box surrounding the point charge
		left = other.x - c_radius;
	   top = other.y + c_radius;
	   right = other.x + c_radius;
	   bottom = other.y - c_radius;

	// Check to see if the particle moved entirely on one side of the charge
		if  ( ((pixel.x < left) && (newx < left))		// moved on left side
	   		||
	         ((pixel.x > right) && (newx > right))		// moved on right side
	         ||
	         ((pixel.y > top) && (newy > top))			// moved on top
	         ||
	         ((pixel.y < bottom) && (newy < bottom))	// moved on bottom
	    	 )
	   {

		// Particles bounce within the bounds of the screen
/*      	if ( (-XBORDER < newx) && (newx < XBORDER) )
         	pixel.x = newx;
         else
         	pixel.sx *= -1;

         if ( (-YBORDER < newy) && (newy < YBORDER) )
         	pixel.y = newy;
         else
         	pixel.sy *= -1;
*/
			pixel.x = newx;
        	pixel.y = newy;
         intersect = FALSE;
	   }

		if (intersect)
	   {
	   	dist = get_distance(pixel, other);
		   if (dist == -1)
		   	cout << "minus one" << endl;

		   // Intersection
		   if (c_radius * c_radius >= dist)
	      {
				intersect = TRUE;
				pixel.color = other.color;
	         return;
	      }
		   else
		   {
		   	intersect = FALSE;
      	if ( (-XBORDER < newx) && (newx < XBORDER) )
         	pixel.x = newx;
         else
         	pixel.sx *= -1;

         if ( (-YBORDER < newy) && (newy < YBORDER) )
         	pixel.y = newy;
         else
         	pixel.sy *= -1;
			}

		}
	}
	return;
}


void process_particle(Particle& pixel, Particle charge[ ], int many_particles)
{
	int l;
   double distx, disty, distance;
   double electrical_force, acceleration;
   double speed, factor;
   Particle other;

   // Distances
	for (l = 0; l < many_particles; l++)
	{
   	other = charge[l];
	   distx = other.x - pixel.x;
	   disty = other.y - pixel.y;
	   distance = sqrt(distx * distx + disty * disty);

      if (distance == 0)
      	return;

	   electrical_force = CHARGE_FACTOR *
	                        (-pixel.charge * other.charge) /
	                        (distance * distance);

	   acceleration = electrical_force / pixel.mass;

		// Alter the speeds based on the acceleration:
		pixel.sx += acceleration * (distx / distance);
		pixel.sy += acceleration * (disty / distance);

		// Make sure that the speed doesn't go over the speed limits!
		speed = (pixel.sx * pixel.sx
				   +
               pixel.sy * pixel.sy
              );

		if (speed > (LIMIT * LIMIT))
		{
			factor = LIMIT / speed;
			pixel.sx *= factor;
			pixel.sy *= factor;
		}
   }

}


double get_distance(Particle pixel, Particle other)
{
	double newx = pixel.x + pixel.sx;
   double newy = pixel.y + pixel.sy;
   double slope, other_slope;
   double x, y;			// Coordinates of intersection of lines
   // Distance from center of charge to point of intersection
   double distx, disty;

	// Line is either vertical or horizontal
	if (pixel.x == newx)
   	return abs(other.x - pixel.x);
   if (pixel.y == newy)
   	return abs(other.y - pixel.y);

  	slope = (newy - pixel.y) / (newx - pixel.x);

   // Calculate negative recriprocal for perpendicular intersecting line
   other_slope = - 1 / slope;

   x = (slope * pixel.x + other.y - pixel.y - other_slope * other.x) / (slope - other_slope);
   y = other_slope * x - other_slope * other.x + other.y;

	distx = (x - other.x);
   disty = (y - other.y);

   return (distx*distx + disty*disty);
}


Particle assign_defaults(int i, int j)
{
	Particle answer;
	answer.x = xpixel_to_cart(i);
	answer.y = ypixel_to_cart(j);
	answer.charge = 1;					// Positron
	answer.sx = 0;
	answer.sy = 0;
	answer.mass = E_MASS;				// same mass as an electron
	return answer;
}


double xpixel_to_cart(int x)
{
	double percent;

   percent = (double)x / (double)WIDTH;
   return (-XBORDER + percent * 2.0 * XBORDER);
}


double ypixel_to_cart(int y)
{
	double percent;

   percent = (double)y / (double)HEIGHT;
   return (YBORDER - percent * 2.0 * YBORDER);
}



int xtranslate (double x)
{
   double percent;		// Percent of x-hemisphere (positive or negative) where
   							// user defined coordinate is located
   int x_value;			// Corresponding graphical x-value

   percent = x / XBORDER;	// Finds, percentagewise, how far the user defined
   								// coordinate is located from the y-axis.

   x_value = WIDTH/2 + percent * (WIDTH/2);

   return (int)x_value;
}


int ytranslate (double y)
{
   double percent;	// Percent of y-hemisphere (positive or negative) where
   						// user defined coordinate is located
   int y_value;		// Corresponding graphical y-value

   percent = -y / YBORDER;	// Finds, percentagewise, how far the user defined
   								// coordinate is located from the x-axis.  Since the
                           // y-values run opposite from expected, the negative
                           // sign is necessary.

  y_value = HEIGHT/2 + percent * (HEIGHT/2);

   return (int)y_value;
}
